from enum import Enum

class ReservedArgKeywords(Enum):
    DATA_IO = 'data_io'
    DEBUG = 'debug'
    DEBUG_SHORT = 'd'
    SYSTEM_TRANSFER = 'system_transfer'
    SYSTEM_TRANSFER_SHORT = 'st'
    SEED = 'seed'
    CONFIG_NAME = 'config_name'
    CONFIG_NAME_SHORT = 'c'
    COPY_CONFIG = 'copy_config'
    COPY_CONFIG_SHORT = 'cc'
    TRAIN = 'train'
    TRAIN_SHORT = 't'
    INFER = 'infer'
    INFER_SHORT = 'i'
    WIPE = 'wipe'
    WIPE_SHORT = 'w'
    GRID_JOB = 'grid_job'
    GRID_JOB_SHORT = 'g'
    VIEW_GRID_JOB = 'view_grid_job'
    VIEW_GRID_JOB_SHORT = 'vg'
    SEQUENTIAL_GRID_JOB = 'sequential_grid_job'
    SEQUENTIAL_GRID_JOB_SHORT = 'sg'
    CALL_OBJECT = 'call_object'
    CALL_OBJECT_SHORT = 'co'
    CALL_OBJECT_ARGS = 'call_object_args'
    CALL_OBJECT_ARGS_SHORT = 'co_args'
    FIND_IMPORT = 'find_import'
    FIND_IMPORT_SHORT = 'fi'
    OVERLEAF_TRANSFER = 'overleaf_transfer'
    OVERLEAF_TRANSFER_SHORT = 'ot'
    TENSORBOARD = 'tensorboard'
    TENSORBOARD_SHORT = 'tb'